#include "PainterTextureSample.h"




////////////////////////////////////////////////////////////////////////////////////////
//
// For ParamMap2 Create an instance of a class derived from ParamMap2UserDlgProc.  This 
// is used to provide special processing of controls in the rollup page.-- NH 
// The original use of this has been replaced by the use of PBAccessor
//
////////////////////////////////////////////////////////////////////////////////////////

PainterDlgProc::~PainterDlgProc() 

{
	if (iPaintButton) ReleaseICustButton(iPaintButton);
	iPaintButton = NULL;
	if (pPainter)
		{
		if (pPainter->InPaintMode() )
			pPainter->EndPaintSession();
		}
	if (tex) tex->painterProc = NULL;
	if (mod) mod->painterProc = NULL;
}


void PainterDlgProc::DeleteThis() 
{ 				
	delete this;
}


BOOL PainterDlgProc::DlgProc(
		TimeValue t,IParamMap2 *map,
		HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	
	switch (msg) 
		{	
		case WM_INITDIALOG:
			{
			InitUI(hWnd);
			EnableUI(FALSE);
			break;
			}
		case WM_DESTROY:
			{
			if (tex) tex->SaveMap();
			DestroyUI(hWnd);
			break;
			}
		case WM_COMMAND:
			switch (LOWORD(wParam)) 
				{

				case IDC_PAINT:
					Paint();
					break;

				case IDC_PAINTOPTIONS:
					PaintOptions();
					break;


				case IDC_APPLY:
					PaintNode();
					if (tex) 
						{
						tex->texHandleValid.Empty();
						tex->DiscardTexHandle();
						tex->NotifyDependents(FOREVER, PART_TEXMAP, REFMSG_CHANGE);
						}
					if (mod) mod->NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
					GetCOREInterface()->RedrawViews(GetCOREInterface()->GetTime());

					break;

				case IDC_FILL:
					if (tex) tex->Fill();
					else if (mod) mod->Fill();
					if (tex) 
						{
						tex->texHandleValid.Empty();
						tex->DiscardTexHandle();
						tex->NotifyDependents(FOREVER, PART_TEXMAP, REFMSG_CHANGE);
						}
					if (mod) mod->NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
					GetCOREInterface()->RedrawViews(GetCOREInterface()->GetTime());
					break;

				}
		}

	return FALSE;
	
}


void  PainterDlgProc::InitUI(HWND hWnd)
	{
	hwnd = hWnd;
	if (iPaintButton) ReleaseICustButton(iPaintButton);
	iPaintButton = NULL;

	iPaintButton = GetICustButton(GetDlgItem(hWnd,IDC_PAINT));
	iPaintButton->SetType(CBT_CHECK);
	iPaintButton->SetHighlightColor(GREEN_WASH);

	pPainter = NULL;

	ReferenceTarget *painterRef = (ReferenceTarget *) GetCOREInterface()->CreateInstance(REF_TARGET_CLASS_ID,PAINTERINTERFACE_CLASS_ID);
	
//set it to the correct verion
	if (painterRef)
		{
		pPainter = (IPainterInterface_V5 *) painterRef->GetInterface(PAINTERINTERFACE_V5);
		}
	if (tex) tex->pPainter = pPainter;
	if (mod) mod->pPainter = pPainter;
	if (tex) tex->painterHwnd = hWnd;


	}
void  PainterDlgProc::DestroyUI(HWND hWnd)
	{
	if (iPaintButton) ReleaseICustButton(iPaintButton);
	iPaintButton = NULL;
	if (pPainter)
		{
		if (pPainter->InPaintMode() )
			pPainter->EndPaintSession();
		}
	}

void PainterDlgProc::EnableUI(BOOL enable)
	{
	HWND  gradHwnd = GetDlgItem(hwnd,IDC_GRADIENT_COMBO);
	HWND  inksHwnd = GetDlgItem(hwnd,IDC_INKS_COMBO);

	ICustButton *iApplyButton = GetICustButton(GetDlgItem(hwnd,IDC_APPLY));
	iApplyButton->Enable(enable);
	ReleaseICustButton(iApplyButton);
	
//	ICustButton *iSplineButton = GetICustButton(GetDlgItem(hwnd,IDC_SPLINE));
//	iSplineButton->Enable(enable);
//	ReleaseICustButton(iSplineButton);

	ICustButton *iFillButton = GetICustButton(GetDlgItem(hwnd,IDC_FILL));
	iFillButton->Enable(enable);
	ReleaseICustButton(iFillButton);


	}



void PainterDlgProc::PaintOptions()
{
	if (pPainter) //need to check this since some one could have taken the painterinterface plugin out
		{
		pPainter->BringUpOptions();
		}

}


//Some helper functions to handle painter UI
void PainterDlgProc::Paint()
{
	INode *node = NULL;
//	pblock->GetValue(pb_node,0,node,FOREVER);

	int nodeCount =  GetCOREInterface()->GetSelNodeCount();
	if (nodeCount > 0)
		node = GetCOREInterface()->GetSelNode(0);

	BOOL bail = FALSE;

	if ((pPainter) && (node))//need to check this since some one could have taken the painterinterface plugin out
		{
		if (!pPainter->InPaintMode())  
			{

			EnableUI(TRUE);

			if (tex) pPainter->InitializeCallback((ReferenceTarget *) tex); //initialize the callback
			else if (mod) pPainter->InitializeCallback((ReferenceTarget *) mod); //initialize the callback

	//we dont use the point gather or normal data
			pPainter->SetEnablePointGather(TRUE);
			pPainter->SetBuildNormalData(FALSE);

//this sends all our dependant nodes to the painter
			Tab<INode*> nodes;
			nodes.Append(1,&node);
			pPainter->InitializeNodes(0, nodes);

		//since we are painting on the final mesh we dont need to do any special handling of nodes

			pPainter->StartPaintSession(); //start the paint session

			if (iPaintButton == NULL)
				InitUI(hwnd);

			iPaintButton->SetCheck(TRUE);
			

//need to build mesh and uvw data 
			ObjectState sos;
			TimeValue t = GetCOREInterface()->GetTime();
			if (node)
				{
				sos = node->EvalWorldState(t);
		
				Mesh *msh = NULL;
				TriObject *collapsedtobj = NULL;

				if (sos.obj)
					{
					if (sos.obj->IsSubClassOf(triObjectClassID))
						{
						TriObject *tobj = (TriObject*)sos.obj;
						msh = &tobj->GetMesh();
						}
//collapse it to a mesh
					else if (sos.obj->CanConvertToType(triObjectClassID))
						{
						collapsedtobj = (TriObject*) sos.obj->ConvertToType(t,triObjectClassID);
						msh = &collapsedtobj->GetMesh();
						}
					if (msh != NULL)
						{

						Matrix3 tm = node->GetObjectTM(t);
						int ct = 0;


						Point3 *loresWorldSpaceList = NULL;
						if (tex)
							{
							loresWorldSpaceList = tex->InitMesh(msh, tm, ct);
							tex->node = node;
							}
						else if (mod)
							{
							VertexPaintData *vmd = (VertexPaintData *) mod->GetLocalDataFromNode(node);
							ct = vmd->vertexPosList.Count();
							loresWorldSpaceList = vmd->vertexPosList.Addr(0);
							}

//						Point3 *loresWorldSpaceList = InitMesh(msh, tm, ct);
						
						if (ct > 0)
							pPainter->LoadCustomPointGather(ct, loresWorldSpaceList, node);
						else bail = TRUE;

						pPainter->UpdateMeshes(TRUE);
						}
				

					if ((collapsedtobj) && (collapsedtobj != sos.obj)) collapsedtobj->DeleteThis();

					}
				}
			}
		else //we are currently in a paint mode so turn it off
			{
			EnableUI(FALSE);
			pPainter->EndPaintSession(); //end the paint session
			iPaintButton->SetCheck(FALSE);
			}

		}

	if (bail) 
		{
		EnableUI(FALSE);
		pPainter->EndPaintSession(); //end the paint session
		iPaintButton->SetCheck(FALSE);
		}

}

BOOL  PainterDlgProc::SystemEndPaintSession()
{
	EnableUI(FALSE);
	if (iPaintButton)
		{
		iPaintButton->SetCheck(FALSE);
		ReleaseICustButton(iPaintButton);
		}
	iPaintButton = NULL;
	return TRUE;
}

float PainterDlgProc::GetStrokeLen()
{
	float strokeLen = 0;
	int ct = pPainter->GetStrokeCount();
	Point3 *wp = pPainter->GetStrokePointWorld();

	for (int i = 1; i < ct; i++)
		{
		strokeLen += Length(wp[i] - wp[i-1]);
		}
	return strokeLen;
}

void PainterDlgProc::PaintLine(IPoint2 mousePosS, IPoint2 mousePosE, INode *node)
{

	pPainter->ClearStroke();

	startLine = mousePosS;
	endLine = mousePosE;
	pPainter->AddToStroke(startLine,FALSE,FALSE);
		//do a sample here
	float len = floor(Length(endLine - startLine));
	int ct = sampleRate;
	Point3 eLine,sLine;
	eLine.x = endLine.x;
	eLine.y = endLine.y;
	eLine.z = 0.0f;
	sLine.x = startLine.x;
	sLine.y = startLine.y;
	sLine.z = 0.0f;

	Point3 vec = Normalize(eLine - sLine) * len/(float)ct;
	IPoint2 currentPos(startLine);

	for (int i = 0; i < ct; i++)
		{
		sLine += vec;
		currentPos.x = sLine.x;
		currentPos.y = sLine.y;

//DebugPrint("pline mouse pos %d %d\n",currentPos.x ,currentPos.y );
		pPainter->AddToStroke(currentPos,FALSE,FALSE);
		}
	pPainter->AddToStroke(endLine,TRUE,TRUE);


	float strokeLen = GetStrokeLen();

	int count = 0;
		
			
	if (tex) 
		{
	
		float *str = pPainter->RetrievePointGatherStr(node, count);
		float *weight = pPainter->RetrievePointGatherWeights(node, count);
		float *u = pPainter->RetrievePointGatherU(node, count);
		tex->inker.ApplyStroke(tex->bm, weight, str, tex->loresUVWIndexList.Addr(0), count ,tex->strokeColor,
			tex->GetGradient(),u,tex->GetGradientTex(),strokeLen);
		}	
	else if (mod)
		{
		for (int i = 0; i < mod->vmdList.Count(); i++)
			{
			float *str = pPainter->RetrievePointGatherStr(mod->vmdList[i]->node, count);
			float *weight = pPainter->RetrievePointGatherWeights(mod->vmdList[i]->node, count);
			float *u = pPainter->RetrievePointGatherU(mod->vmdList[i]->node, count);

			mod->inker.ApplyStroke(mod->vmdList[i]->vcColor.Addr(0), weight, str, count,mod->strokeColor,
					mod->GetGradient(),u,mod->GetGradientTex(),strokeLen);
			}
		}
}

void PainterDlgProc::PaintRectangle(IPoint2 mousePosS, IPoint2 mousePosE, INode *node)
{
	pPainter->ClearStroke();
	startLine = mousePosS;
	endLine = mousePosE;
	for (int j = 0; j < 4; j++)
		{

		IPoint2 startRLine, endRLine;

		if (j==2)
			{
			startRLine.x = startLine.x;
			startRLine.y = startLine.y;
			endRLine.x = endLine.x;
			endRLine.y = startLine.y;
			}
		else if (j==3)
			{
			startRLine.x = endLine.x;
			startRLine.y = startLine.y;
			endRLine.x = endLine.x;
			endRLine.y = endLine.y;
			}
		else if (j==0)
			{
			startRLine.x = endLine.x;
			startRLine.y = endLine.y;
			endRLine.x = startLine.x;
			endRLine.y = endLine.y;
			}
		else if (j==1)
			{
			startRLine.x = startLine.x;
			startRLine.y = endLine.y;
			endRLine.x = startLine.x;
			endRLine.y = startLine.y;
			}


		pPainter->AddToStroke(startRLine,FALSE,FALSE);
			//do a sample here
		float len = floor(Length(endRLine - startRLine));
		int ct = sampleRate;
		Point3 eLine,sLine;

		eLine.x = endRLine.x;
		eLine.y = endRLine.y;
		eLine.z = 0.0f;
		sLine.x = startRLine.x;
		sLine.y = startRLine.y;
		sLine.z = 0.0f;

		Point3 vec = Normalize(eLine - sLine) * len/(float)ct;
		IPoint2 currentPos(startLine);

		for (int i = 0; i < ct; i++)
			{
			sLine += vec;
			currentPos.x = sLine.x;
			currentPos.y = sLine.y;
			pPainter->AddToStroke(currentPos,FALSE,FALSE);
			}
		if (j==3)
			pPainter->AddToStroke(endRLine,TRUE,TRUE);
		else pPainter->AddToStroke(endRLine,FALSE,FALSE);
					
		}

	float strokeLen = GetStrokeLen();


	int count = 0;
		
			
	if (tex) 
		{
	
		float *str = pPainter->RetrievePointGatherStr(node, count);
		float *weight = pPainter->RetrievePointGatherWeights(node, count);
		float *u = pPainter->RetrievePointGatherU(node, count);
		tex->inker.ApplyStroke(tex->bm, weight, str, tex->loresUVWIndexList.Addr(0), count ,tex->strokeColor,
					tex->GetGradient(),u,tex->GetGradientTex(),strokeLen);
		}	
	else if (mod)
		{
		for (int i = 0; i < mod->vmdList.Count(); i++)
			{
			float *str = pPainter->RetrievePointGatherStr(mod->vmdList[i]->node, count);
			float *weight = pPainter->RetrievePointGatherWeights(mod->vmdList[i]->node, count);
			float *u = pPainter->RetrievePointGatherU(mod->vmdList[i]->node, count);
			mod->inker.ApplyStroke(mod->vmdList[i]->vcColor.Addr(0), weight, str, count,mod->strokeColor,
					mod->GetGradient(),u,mod->GetGradientTex(),strokeLen);
			}
		}

}


void PainterDlgProc::PaintCircle(IPoint2 mousePosS, IPoint2 mousePosE, INode *node)
{
	pPainter->ClearStroke();
	startLine = mousePosS;
	endLine = mousePosE;

	int ct = 4 * sampleRate;

	float angle = 0.0f;
	float inc = (PI*2) / (ct-1);
	float radius = Length(startLine-endLine);



	for (int i = 0; i < ct; i++)
		{
		IPoint2 p;
		p.x =  (radius * cos(angle) + radius * sin(angle)) + startLine.x;
		p.y =  (radius * sin(angle) - radius * cos(angle)) + startLine.y;

		if (i == ct-1)
			pPainter->AddToStroke(p,TRUE,TRUE);
		else pPainter->AddToStroke(p,FALSE,FALSE);
		angle += inc;
		}

	float strokeLen = GetStrokeLen();

	int count = 0;
		
	if (tex) 
		{
		float *str = pPainter->RetrievePointGatherStr(node, count);
		float *weight = pPainter->RetrievePointGatherWeights(node, count);
		float *u = pPainter->RetrievePointGatherU(node, count);
		tex->inker.ApplyStroke(tex->bm, weight, str, tex->loresUVWIndexList.Addr(0), count ,tex->strokeColor,
					tex->GetGradient(),u,tex->GetGradientTex(),strokeLen);
		}	
	else if (mod)
		{
		for (int i = 0; i < mod->vmdList.Count(); i++)
			{
			float *str = pPainter->RetrievePointGatherStr(mod->vmdList[i]->node, count);
			float *weight = pPainter->RetrievePointGatherWeights(mod->vmdList[i]->node, count);
			float *u = pPainter->RetrievePointGatherU(mod->vmdList[i]->node, count);
			mod->inker.ApplyStroke(mod->vmdList[i]->vcColor.Addr(0), weight, str, count,mod->strokeColor,
					mod->GetGradient(),u,mod->GetGradientTex(),strokeLen);
			}
		}


}

void PainterDlgProc::PaintNode()

{

	Interval valid;
	TimeValue t = GetCOREInterface()->GetTime();

	INode *node = NULL;
	if (tex) node = tex->node;

	INode *meshNode = NULL;
	
	if (tex) meshNode = tex->GetNode();
	else if (mod) meshNode = mod->GetNode();

	if (meshNode == NULL) return;
//getnode


	if (tex) 
		tex->HoldMap();
	else if (mod)
		mod->HoldMap();
	theHold.Accept(GetString(IDS_PW_PAINT));

	Matrix3 tm = meshNode->GetObjectTM(t,&valid);
	ObjectState nos = meshNode->EvalWorldState(t);

	if (tex) 
		{
		sampleRate = tex->GetSampleRate();
		}
	else if (mod) 
		{
		sampleRate = mod->GetSampleRate();
		}

	ViewExp *vpt = GetCOREInterface()->GetActiveViewport();
	GraphicsWindow *gw = vpt->getGW();
//	vpt->GetAffineTM(tm);
	
	gw->setTransform(Matrix3(1));


	float strokeLen = GetStrokeLen();

	if (nos.obj->IsShapeObject()) 
		{
		ShapeObject *pathOb = (ShapeObject*)nos.obj;
		if (pathOb->NumberOfCurves()) 
			{
//sample down the spline
			for (int i = 0; i < pathOb->NumberOfCurves(); i++)
				{
				int ct = pathOb->NumberOfVertices(t,i) * sampleRate;
				float u = 0;
				float inc = 1.0f/(ct-1);
				Tab<Point3> plist;
				plist.SetCount(ct);
				for (int j = 0; j < ct; j++)
					{
					plist[j] = pathOb->InterpCurve3D(t, i, u,PARAM_NORMALIZED )*tm;
//convert it to screen space
					Point3 inPoint,outPoint;
					inPoint = plist[j];
					DWORD flag = gw->transPoint(&inPoint, &outPoint);
					plist[j] = outPoint ;
					plist[j].z = 0.0f;

					u += inc;
//DebugPrint("spline %f  mouse pos %d %d\n",u,(int)plist[j].x,(int)plist[j].y);
					}
//clear the stroke
				pPainter->ClearStroke();
				
//apply the stroke
				for (j = 0; j < ct; j++)
					{
					IPoint2 p;
					p.x = (int) plist[j].x;
					p.y = (int) plist[j].y;
					if (j == ct-1)
						pPainter->AddToStroke(p,TRUE,TRUE);
					else pPainter->AddToStroke(p,FALSE,FALSE);
					}

//save the bitmap
				float strokeLen = GetStrokeLen();

				int count = 0;
				if (tex) 
					{
					float *str = pPainter->RetrievePointGatherStr(node, count);
					float *weight = pPainter->RetrievePointGatherWeights(node, count);
					float *u = pPainter->RetrievePointGatherU(node, count);
					tex->inker.ApplyStroke(tex->bm, weight, str, tex->loresUVWIndexList.Addr(0), count ,tex->GetStrokeColor(),
								tex->GetGradient(),u,tex->GetGradientTex(),strokeLen);

					int ct = pPainter->GetStrokeCount();
					str = pPainter->GetStrokeStr();
					float *radius = pPainter->GetStrokeRadius();
					Point3 *point = pPainter->GetStrokePointWorld();
					Point3 *mirrorPoint = pPainter->GetStrokePointWorldMirror();

					if (mirrorPoint)
						tex->PaintHiresStroke(ct, radius, str, point, mirrorPoint);
					else tex->PaintHiresStroke(ct, radius, str, point, NULL);					
					}	
				else if (mod)
					{
					for (int k = 0; k < mod->vmdList.Count(); k++)
						{
						float *str = pPainter->RetrievePointGatherStr(mod->vmdList[k]->node, count);
						float *weight = pPainter->RetrievePointGatherWeights(mod->vmdList[k]->node, count);
						float *u = pPainter->RetrievePointGatherU(mod->vmdList[k]->node, count);
						mod->inker.ApplyStroke(mod->vmdList[k]->vcColor.Addr(0), weight, str, count,mod->GetStrokeColor(),
							mod->GetGradient(),u,mod->GetGradientTex(),strokeLen);
						}
					}
				}
			}
		}

//check if spline
//mesh
//else just use the transform
	GetCOREInterface()->ReleaseViewport(vpt);

}



BOOL   PainterDlgProc::StartStroke()
{
	lagRate = pPainter->GetLagRate();
	lagCount = 0;
	firstLineStroke = TRUE;


	if (tex) 
		{
		strokeType = tex->GetStrokeType();
		sampleRate = tex->GetSampleRate();
		}
	else if (mod) 
		{
		strokeType = mod->GetStrokeType();
		sampleRate = mod->GetSampleRate();
		}

	return TRUE;
}
		//This is called as the user strokes across the mesh or screen with the mouse down
BOOL  PainterDlgProc::PaintStroke(
						  BOOL hit,
						  IPoint2 mousePos, 
						  Point3 worldPoint, Point3 worldNormal,
						  Point3 localPoint, Point3 localNormal,
						  Point3 bary,  int index,
						  BOOL shift, BOOL ctrl, BOOL alt, 
						  float radius, float str,
						  float pressure, INode *node,
						  BOOL mirrorOn,
						  Point3 worldMirrorPoint, Point3 worldMirrorNormal,
						  Point3 localMirrorPoint, Point3 localMirrorNormal
						  ) 
{

	
	if (hit)
		{

		float strokeLen = GetStrokeLen();

		if (strokeType == 0) //freeform
			{
			int count = 0;


			if (tex) 
				{
				float *str = pPainter->RetrievePointGatherStr(node, count);
				float *weight = pPainter->RetrievePointGatherWeights(node, count);
				float *u = pPainter->RetrievePointGatherU(node, count);
				tex->inker.ApplyStroke(tex->bm, weight, str, tex->loresUVWIndexList.Addr(0), count ,tex->strokeColor,
					tex->GetGradient(),u,tex->GetGradientTex(),strokeLen);
				}
			else if (mod)
				{
				for (int i = 0; i < mod->vmdList.Count(); i++)
					{
					float *str = pPainter->RetrievePointGatherStr(mod->vmdList[i]->node, count);
					float *weight = pPainter->RetrievePointGatherWeights(mod->vmdList[i]->node, count);
					float *u = pPainter->RetrievePointGatherU(mod->vmdList[i]->node, count);
					mod->inker.ApplyStroke(mod->vmdList[i]->vcColor.Addr(0), weight, str, count,mod->strokeColor,
									mod->GetGradient(),u,mod->GetGradientTex(),strokeLen);
					}
				}

			}
		else if (strokeType == 1) //straight line
			{
			
			if (firstLineStroke)
				{
				startLine = mousePos;
				pPainter->AddToStroke(startLine,FALSE,FALSE);
				firstLineStroke = FALSE;
				}
			else 
				{
				endLine = mousePos;
				PaintLine(startLine,endLine,node);
				}

			}
		else if (strokeType == 2) //rectangle line
			{
			if (firstLineStroke)
				{
				startLine = mousePos;
				pPainter->AddToStroke(startLine,FALSE,FALSE);
				firstLineStroke = FALSE;
				}
			else 
				{
				endLine = mousePos;			
				PaintRectangle(startLine,endLine,node);
				}

			}
		else if (strokeType == 3) //circle line
			{
			if (firstLineStroke)
				{
				startLine = mousePos;
				pPainter->AddToStroke(startLine,FALSE,FALSE);
				firstLineStroke = FALSE;
				}
			else 
				{
				endLine = mousePos;			
				PaintCircle(startLine,endLine,node);
				}

			}
		}


	lagCount++;
	if ((lagRate == 0) || ((lagCount%lagRate) == 0))
		{
		if (tex)
			{
			tex->texHandleValid.Empty();
			tex->DiscardTexHandle();
			}
		else if (mod)
			{
			mod->NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
			}
		}

	return TRUE;
}

		// This is called as the user ends a strokes when the users has it set to always update
BOOL  PainterDlgProc::EndStroke()
{
return TRUE;
}

		// This is called as the user ends a strokes when the users has it set to update on mouse up only
		// the canvas gets a list of all points, normals etc instead of one at a time
		//		int ct - the number of elements in the following arrays
		//  <...> see paintstroke() these are identical except they are arrays of values
BOOL  PainterDlgProc::EndStroke(int ct, BOOL *hit, IPoint2 *mousePos, 
						  Point3 *worldPoint, Point3 *worldNormal,
						  Point3 *localPoint, Point3 *localNormal,
						  Point3 *bary,  int *index,
						  BOOL *shift, BOOL *ctrl, BOOL *alt, 
						  float *radius, float *str,
						  float *pressure, INode **node,
						  BOOL mirrorOn,
						  Point3 *worldMirrorPoint, Point3 *worldMirrorNormal,
						  Point3 *localMirrorPoint, Point3 *localMirrorNormal)
{

	if (strokeType == 0)
		{
		int count = 0;
		float *str = NULL;
		float *weight = NULL;

		float strokeLen = GetStrokeLen();
		if (tex) 
			{
			str = pPainter->RetrievePointGatherStr(tex->node, count);
			weight = pPainter->RetrievePointGatherWeights(tex->node, count);
			float *u = pPainter->RetrievePointGatherU(tex->node, count);

			tex->inker.ApplyStroke(tex->bm, weight, str, tex->loresUVWIndexList.Addr(0), count, tex->strokeColor,
									tex->GetGradient(),u,tex->GetGradientTex(),strokeLen);
			}
		else if (mod)
			{
			for (int i = 0; i < mod->vmdList.Count(); i++)
				{
				float *str = pPainter->RetrievePointGatherStr(mod->vmdList[i]->node, count);
				float *weight = pPainter->RetrievePointGatherWeights(mod->vmdList[i]->node, count);
				float *u = pPainter->RetrievePointGatherU(mod->vmdList[i]->node, count);

				mod->inker.ApplyStroke(mod->vmdList[i]->vcColor.Addr(0), weight, str, count,mod->strokeColor,
									mod->GetGradient(),u,mod->GetGradientTex(),strokeLen);
				}
			}
		}
	else if (strokeType == 1)
		{
		endLine = mousePos[0];
		startLine = mousePos[ct-1];
		if (tex)
			PaintLine(endLine,startLine,tex->node);
		else PaintLine(endLine,startLine,NULL);
		}
	else if (strokeType == 2)
		{
		endLine = mousePos[0];
		startLine = mousePos[ct-1];
		if (tex)
			PaintRectangle(endLine,startLine,tex->node);
		else PaintRectangle(endLine,startLine,NULL);		
		}
	else if (strokeType == 3)
		{
		endLine = mousePos[0];
		startLine = mousePos[ct-1];
		if (tex)
			PaintCircle(endLine,startLine,tex->node);
		else PaintCircle(endLine,startLine,NULL);		
		}

	return TRUE;
}

		// This is called as the user cancels a stroke by right clicking
BOOL  PainterDlgProc::CancelStroke()
{
return TRUE;
}



void PainterDlgProc::PainterDisplay(TimeValue t, ViewExp *vpt, int flags) 

{

	int ct = pPainter->GetStrokeCount();
	IPoint2 *mousePos = pPainter->GetStrokeMousePos();

	IPoint2 endLine = mousePos[ct-1];
	IPoint2 startLine = mousePos[0];

	if (tex) 
		strokeType = tex->GetStrokeType();
	else if (mod) 
		strokeType = mod->GetStrokeType();

	if (tex) 
		sampleRate = tex->GetSampleRate();
	else if (mod) 
		sampleRate = mod->GetSampleRate();

	if (strokeType == 1)
		{
		GraphicsWindow *gw = vpt->getGW();

		Matrix3 tm(1);
		gw->setTransform(tm);

		int savedLimits;
		gw->setRndLimits((savedLimits = gw->getRndLimits()) & (~GW_ILLUM) );
		IPoint3 line[3];
		line[0].x = startLine.x;
		line[0].y = startLine.y;
		line[0].z = 0;

		line[1].x = endLine.x;
		line[1].y = endLine.y;
		line[1].z = 0;

		gw->wPolyline(2, line, NULL, FALSE, NULL);

		gw->setRndLimits(savedLimits);
		}
	else if (strokeType == 2)
		{
		GraphicsWindow *gw = vpt->getGW();
		Matrix3 tm(1);
		gw->setTransform(tm);
		int savedLimits;
		gw->setRndLimits((savedLimits = gw->getRndLimits()) & (~GW_ILLUM) );

		IPoint3 line[6];
/*
		line[0].x = startLine.x;
		line[0].y = startLine.y;
		line[0].z = 0;

		line[1].x = endLine.x;
		line[1].y = endLine.y;
		line[1].z = 0;

		gw->wPolyline(2, line, NULL, FALSE, NULL);
*/

		line[0].x = startLine.x;
		line[0].y = startLine.y;
		line[0].z = 0;

		line[1].x = endLine.x;
		line[1].y = startLine.y;
		line[1].z = 0;

		line[2].x = endLine.x;
		line[2].y = endLine.y;
		line[2].z = 0;

		line[3].x = startLine.x;
		line[3].y = endLine.y;
		line[3].z = 0;

		line[4].x = startLine.x;
		line[4].y = startLine.y;
		line[4].z = 0;

		gw->wPolyline(5, line, NULL, FALSE, NULL);

		gw->setRndLimits(savedLimits);
		}
	else if (strokeType == 3)
		{

		GraphicsWindow *gw = vpt->getGW();
		int savedLimits;

		Matrix3 tm(1);
		gw->setTransform(tm);

		gw->setRndLimits((savedLimits = gw->getRndLimits()) & (~GW_ILLUM) );



		IPoint3 line[6];
		int ct = 4 * sampleRate;

		float angle = 0.0f;
		float inc = (PI*2) / (ct-1);
		float radius = Length(startLine-endLine);



		for (int i = 0; i < (ct-1); i++)
			{
			line[0].x =  (radius * cos(angle) + radius * sin(angle)) + startLine.x;
			line[0].y =  (radius * sin(angle) - radius * cos(angle)) + startLine.y;

			line[1].x =  (radius * cos(angle+inc) + radius * sin(angle+inc)) + startLine.x;
			line[1].y =  (radius * sin(angle+inc) - radius * cos(angle+inc)) + startLine.y;

			gw->wPolyline(2, line, NULL, FALSE, NULL);

			angle += inc;
			}

		gw->setRndLimits(savedLimits);

		}
}
